import type Firebase from 'firebase';
export declare const ModuleName = "firestorter";
export interface IContext {
    readonly firebase: typeof Firebase;
    readonly app: Firebase.app.App;
    readonly firestore: Firebase.firestore.Firestore;
}
export interface IHasContext {
    readonly context?: IContext;
}
export declare type FirestorterConfig = {
    firebase: typeof Firebase;
    app?: string | Firebase.app.App;
    firestore?: Firebase.firestore.Firestore;
};
/**
 * Initializes `firestorter` with the firebase-app.
 *
 * @param {Object} config - Configuration options
 * @param {Firebase} config.firebase - Firebase reference
 * @param {String|FirebaseApp} [config.app] - FirebaseApp to use (when omitted the default app is used)
 *
 * @example
 * import * as firebase from 'firebase/app';
 * import 'firebase/firestore';
 * import {initFirestorter, Collection, Document} from 'firestorter';
 *
 * // Initialize firebase app
 * firebase.initializeApp({...});
 *
 * // Initialize `firestorter`
 * initFirestorter({firebase: firebase});
 *
 * // Create collection or document
 * const albums = new Collection('artists/Metallica/albums');
 * ...
 * const album = new Document('artists/Metallica/albums/BlackAlbum');
 * ...
 */
declare function initFirestorter(config: FirestorterConfig): void;
/**
 * If you need to use different firestore instances for different
 * collections, or otherwise want to avoid global state, you can
 * instead provide a "context" option when creating Document and
 * Collection instances.
 *
 * This function takes the same arguments as `initFirestorter` and returns
 * a context suitable for Document and Collection creation.
 *
 * @example
 * import * as firebase from 'firebase/app';
 * import 'firebase/firestore'
 * import * as firetest from '@firebase/testing'
 * import { makeFirestorterContext, Collection, Document } from "firestorter"
 *
 * function makeTestContext(fbtestArgs) {
 * 	 const app = firetest.initializeTestApp(fbtestArgs)
 *   return makeFirestorterContext({
 *     firestore,
 *     app,
 *   })
 * }
 *
 * // create collection or document without global state
 * test('collection and document using different apps', () => {
 *   const context1 = makeTestContext({ projectId: 'foo' })
 *   const context2 = makeTestContext({ projectId: 'bar' })
 *
 *   // Create collection or document
 *   const albums = new Collection('artists/Metallica/albums', {context: context1});
 *   ...
 *   const album = new Document('artists/Metallica/albums/BlackAlbum', {context: context2});
 *   ...
 * })
 */
export declare function makeFirestorterContext(config: FirestorterConfig): IContext;
declare function getFirebase(obj?: IHasContext): typeof Firebase;
declare function getFirebaseApp(obj?: IHasContext): Firebase.app.App;
declare function getFirestore(obj?: IHasContext): Firebase.firestore.Firestore;
export { initFirestorter, getFirestore, getFirebase, getFirebaseApp };
